Package: rtide
Title: Tide Heights
Version: 0.0.12
Authors@R: c(
    person("Joe", "Thorley", , "joe@poissonconsulting.ca", role = "aut",
           comment = c(ORCID = "0000-0002-7683-4592")),
    person("Luke", "Miller", , "contact@lukemiller.org", role = c("aut", "cre")),
    person("Abram", "Fleishman", role = "aut"),
    person("Poisson Consulting", role = "cph")
  )
Description: Calculates tide heights based on tide station harmonics.  It
    includes the harmonics data for 637 US stations.  The harmonics data
    was converted from
    <https://github.com/poissonconsulting/rtide/blob/main/data-raw/harmonics-dwf-20151227-free.tar.bz2>,
    NOAA web site data processed by David Flater for 'XTide'.  The code to
    calculate tide heights from the harmonics is based on 'XTide'.
License: GPL-3
URL: https://github.com/millerlp/rtide
BugReports: https://github.com/millerlp/rtide/issues
Depends: R (>= 4.0)
Imports: abind, chk (>= 0.10.0), dttr2, tibble, utils
Suggests: covr, ggplot2, scales, spelling, testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-25 04:25:02 UTC; millerlp
Author: Joe Thorley [aut] (ORCID: <https://orcid.org/0000-0002-7683-4592>),
  Luke Miller [aut, cre],
  Abram Fleishman [aut],
  Poisson Consulting [cph]
Maintainer: Luke Miller <contact@lukemiller.org>
Repository: CRAN
Date/Publication: 2026-07-25 04:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(format,tide_harmonics)
S3method(print,tide_harmonics)
S3method(subset,tide_harmonics)
export(is.tide_harmonics)
export(tide_datetimes)
export(tide_height)
export(tide_height_data)
export(tide_slack_data)
export(tide_stations)
import(chk)
import(dttr2)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rtide

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![R-CMD-check](https://github.com/millerlp/rtide/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/millerlp/rtide/actions/workflows/R-CMD-check.yaml)
<!--[![Codecov test coverage](https://codecov.io/gh/poissonconsulting/rtide/graph/badge.svg)](https://app.codecov.io/gh/poissonconsulting/rtide) -->
[![License:
GPL3](https://img.shields.io/badge/License-GPL3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0.en.html)
[![CRAN
status](https://www.r-pkg.org/badges/version/rtide)](https://cran.r-project.org/package=rtide)
![CRAN downloads](https://cranlogs.r-pkg.org/badges/rtide) [![Codecov
test
coverage](https://codecov.io/gh/millerlp/rtide/graph/badge.svg)](https://app.codecov.io/gh/millerlp/rtide)
<!-- badges: end -->

## Introduction

`rtide` is an R package to calculate tide heights based on tide station
harmonics.

It includes the harmonics data for 637 US stations.

## Installation

To install the latest release from [CRAN](https://cran.r-project.org)

``` r
install.packages("rtide")
```

To install the developmental version from
[GitHub](https://github.com/millerlp/rtide)

``` r
# install.packages("pak")
pak::pak("millerlp/rtide")
```

## Utilisation

``` r
library(tibble)
library(rtide)
#> rtide is not suitable for navigation

data <- rtide::tide_height(
  "Monterey Harbor",
  from = as.Date("2016-07-13"), to = as.Date("2016-07-15"),
  minutes = 10L, tz = "America/Los_Angeles"
)

print(data)
#> # A tibble: 432 × 3
#>    Station                               DateTime            TideHeight
#>    <chr>                                 <dttm>                   <dbl>
#>  1 Monterey, Monterey Harbor, California 2016-07-13 00:00:00      0.514
#>  2 Monterey, Monterey Harbor, California 2016-07-13 00:10:00      0.496
#>  3 Monterey, Monterey Harbor, California 2016-07-13 00:20:00      0.481
#>  4 Monterey, Monterey Harbor, California 2016-07-13 00:30:00      0.468
#>  5 Monterey, Monterey Harbor, California 2016-07-13 00:40:00      0.457
#>  6 Monterey, Monterey Harbor, California 2016-07-13 00:50:00      0.449
#>  7 Monterey, Monterey Harbor, California 2016-07-13 01:00:00      0.443
#>  8 Monterey, Monterey Harbor, California 2016-07-13 01:10:00      0.440
#>  9 Monterey, Monterey Harbor, California 2016-07-13 01:20:00      0.439
#> 10 Monterey, Monterey Harbor, California 2016-07-13 01:30:00      0.441
#> # ℹ 422 more rows
```

``` r
library(ggplot2)
library(scales)
```

``` r
ggplot(data = data, aes(x = DateTime, y = TideHeight)) +
  geom_line() +
  scale_x_datetime(
    name = "Date",
    labels = date_format("%d %b %Y", tz = "America/Los_Angeles")
  ) +
  scale_y_continuous(name = "Tide Height (m)") +
  ggtitle("Monterey Harbour")
```

![](man/figures/README-unnamed-chunk-6-1.png)<!-- -->

## Shiny

Tide heights can be also obtained using rtide through a [shiny
interface](https://poissonconsulting.shinyapps.io/rtide/) developed by
Seb Dalgarno.

## Contribution

Please report any [issues](https://github.com/millerlp/rtide/issues).

[Pull requests](https://github.com/millerlp/rtide/pulls) are always
welcome.

<!--  ## Code of Conduct
  &#10;Please note that the rtide project is released with a [Contributor Code of Conduct](https://poissonconsulting.github.io/rtide/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms. -->

## Inspiration

The harmonics data was converted from
<https://github.com/poissonconsulting/rtide/blob/main/data-raw/harmonics-dwf-20151227-free.tar.bz2>,
NOAA web site data processed by David Flater for
[`XTide`](https://flaterco.com/xtide/). The code to calculate tide
heights from the harmonics is based on `XTide`.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rtide

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![R-CMD-check](https://github.com/millerlp/rtide/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/millerlp/rtide/actions/workflows/R-CMD-check.yaml)
<!--[![Codecov test coverage](https://codecov.io/gh/poissonconsulting/rtide/graph/badge.svg)](https://app.codecov.io/gh/poissonconsulting/rtide) -->
[![License:
GPL3](https://img.shields.io/badge/License-GPL3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0.en.html)
[![CRAN
status](https://www.r-pkg.org/badges/version/rtide)](https://cran.r-project.org/package=rtide)
![CRAN downloads](https://cranlogs.r-pkg.org/badges/rtide)
<!-- badges: end -->

## Introduction

`rtide` is an R package to calculate tide heights based on tide station
harmonics.

It includes the harmonics data for 637 US stations.

## Installation

To install the latest release from [CRAN](https://cran.r-project.org)

``` r
install.packages("rtide")
```

To install the developmental version from
[GitHub](https://github.com/millerlp/rtide)

``` r
# install.packages("pak")
pak::pak("millerlp/rtide")
```

## Utilisation

``` r
library(tibble)
library(rtide)
#> rtide is not suitable for navigation

data <- rtide::tide_height(
  "Monterey Harbor",
  from = as.Date("2016-07-13"), to = as.Date("2016-07-15"),
  minutes = 10L, tz = "America/Los_Angeles"
)

print(data)
#> # A tibble: 432 × 3
#>    Station                               DateTime            TideHeight
#>    <chr>                                 <dttm>                   <dbl>
#>  1 Monterey, Monterey Harbor, California 2016-07-13 00:00:00      0.514
#>  2 Monterey, Monterey Harbor, California 2016-07-13 00:10:00      0.496
#>  3 Monterey, Monterey Harbor, California 2016-07-13 00:20:00      0.481
#>  4 Monterey, Monterey Harbor, California 2016-07-13 00:30:00      0.468
#>  5 Monterey, Monterey Harbor, California 2016-07-13 00:40:00      0.457
#>  6 Monterey, Monterey Harbor, California 2016-07-13 00:50:00      0.449
#>  7 Monterey, Monterey Harbor, California 2016-07-13 01:00:00      0.443
#>  8 Monterey, Monterey Harbor, California 2016-07-13 01:10:00      0.440
#>  9 Monterey, Monterey Harbor, California 2016-07-13 01:20:00      0.439
#> 10 Monterey, Monterey Harbor, California 2016-07-13 01:30:00      0.441
#> # ℹ 422 more rows
```

``` r
library(ggplot2)
library(scales)
```

``` r
ggplot(data = data, aes(x = DateTime, y = TideHeight)) +
  geom_line() +
  scale_x_datetime(
    name = "Date",
    labels = date_format("%d %b %Y", tz = "America/Los_Angeles")
  ) +
  scale_y_continuous(name = "Tide Height (m)") +
  ggtitle("Monterey Harbour")
```

![](man/figures/README-unnamed-chunk-6-1.png)<!-- -->

## Shiny

Tide heights can be also obtained using rtide through a [shiny
interface](https://poissonconsulting.shinyapps.io/rtide/) developed by
Seb Dalgarno.

## Contribution

Please report any [issues](https://github.com/millerlp/rtide/issues).

[Pull requests](https://github.com/millerlp/rtide/pulls) are always
welcome.

<!--  ## Code of Conduct
  &#10;Please note that the rtide project is released with a [Contributor Code of Conduct](https://poissonconsulting.github.io/rtide/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms. -->

## Inspiration

The harmonics data was converted from
<https://github.com/poissonconsulting/rtide/blob/main/data-raw/harmonics-dwf-20151227-free.tar.bz2>,
NOAA web site data processed by David Flater for
[`XTide`](https://flaterco.com/xtide/). The code to calculate tide
heights from the harmonics is based on `XTide`.

tide_slack_datetime <- function(d, h, high = TRUE, forward = TRUE) {
  minutes <- seq(0L, 375L, by = 15L)
  d <- rep(d, length(minutes))
  if (!forward) minutes <- minutes * -1L
  d <- dtt_add_minutes(d, minutes)
  height <- vapply(d, tide_height_datetime, 1, h = h)
  which <- which.max(height * if (high) 1 else -1)
  d <- d[which]

  minutes <- -15:15
  d <- rep(d, length(minutes))
  d <- dtt_add_minutes(d, minutes)
  height <- vapply(d, tide_height_datetime, 1, h = h)
  which <- which.max(height * if (high) 1 else -1)
  d <- d[which]

  seconds <- seq(-30L, 30L, by = 3L)
  d <- rep(d, length(seconds))
  d <- dtt_add_seconds(d, seconds)
  height <- vapply(d, tide_height_datetime, 1, h = h)
  which <- which.max(height * if (high) 1 else -1)
  d <- d[which]

  seconds <- -3:3
  d <- rep(d, length(seconds))
  d <- dtt_add_seconds(d, seconds)
  height <- vapply(d, tide_height_datetime, 1, h = h)
  which <- which.max(height * if (high) 1 else -1)
  d[which]
}

tide_slack_data_datetime <- function(d, h) {
  datetimes <- list(
    tide_slack_datetime(d$DateTime, h, TRUE, TRUE),
    tide_slack_datetime(d$DateTime, h, TRUE, FALSE),
    tide_slack_datetime(d$DateTime, h, FALSE, TRUE),
    tide_slack_datetime(d$DateTime, h, FALSE, FALSE)
  )

  seconds <- vapply(datetimes, datetime2seconds, 1)
  which <- which.min(abs(seconds - datetime2seconds(d$DateTime)))

  d$SlackDateTime <- datetimes[[which]]
  d$SlackTideHeight <- tide_height_datetime(d$SlackDateTime, h = h)
  d$SlackType <- if (which %in% 1:2) "high" else "low"
  d
}

tide_slack_data_station <- function(data, harmonics) {
  harmonics <- subset(harmonics, paste0("^", data$Station[1], "$"))
  data <- split(data, seq_len(nrow(data)))
  data <- lapply(data, FUN = tide_slack_data_datetime, h = harmonics)
  data <- do.call("rbind", data)
  if (harmonics$Station$Units %in% c("feet", "ft")) {
    data$SlackTideHeight <- ft2m(data$SlackTideHeight)
  }
  data
}

#' Tide Slack Data
#'
#' Determines the closest slack tide for specified stations at particular date times based on the supplied harmonics object.
#'
#' @param data A data frame with the columns Station and DateTime.
#' @inheritParams tide_stations
#' @return A data frame of the slack tide date times and heights in m.
#' @export
tide_slack_data <- function(data, harmonics = rtide::harmonics) {
  check_data(data,
    values = list(Station = "", DateTime = Sys.time()),
    nrow = c(1L, .Machine$integer.max)
  )

  if (!all(data$Station %in% tide_stations(harmonics = harmonics))) {
    stop("unrecognised stations", call. = FALSE)
  }

  if (has_name(data, "SlackTideHeight")) {
    stop("data already has 'SlackTideHeight' column", call. = FALSE)
  }

  if (has_name(data, "SlackDateTime")) {
    stop("data already has 'SlackDateTime' column", call. = FALSE)
  }

  if (has_name(data, "SlackType")) {
    stop("data already has 'SlackType' column", call. = FALSE)
  }

  tz <- dtt_tz(data$DateTime)
  data$DateTime <- dtt_adjust_tz(data$DateTime, tz = "UTC")

  years <- range(dtt_year(data$DateTime), na.rm = TRUE)
  if (!all(years %in% years_tide_harmonics(harmonics))) {
    stop("years are outside harmonics range", call. = FALSE)
  }

  data <- split(data, data$Station)
  data <- lapply(data, FUN = tide_slack_data_station, harmonics = harmonics)
  op <- options(stringsAsFactors = FALSE)
  on.exit(options(op))
  data <- do.call("rbind", data)

  data$DateTime <- dtt_adjust_tz(data$DateTime, tz = tz)
  data$SlackDateTime <- dtt_adjust_tz(data$SlackDateTime, tz = tz)
  data <- data[order(data$Station, data$DateTime), ]
  tibble::as_tibble(data)
}

test_that("tide_slack_data works", {
  data <- data.frame(
    Station = "Monterey, Monterey Harbor, California",
    DateTime = ISOdate(2015, 1, 1, 10, tz = "PST8PDT"),
    stringsAsFactors = FALSE
  )

  expect_error(chk::check_data(tide_slack_data(data),
    values = list(
      Station = "", DateTime = Sys.time(), SlackDateTime = Sys.time(),
      SlackTideHeight = 1, SlackType = ""
    ),
    nrow = 1, exclusive = TRUE, order = TRUE
  ), NA)
})

test_that("tide_height_data predictions", {
  slack <- tide_slack_data(rtide::monterey)
  expect_equal(slack$MLLW, slack$SlackTideHeight, tolerance = 0.002)
  expect_equal(slack$DateTime, slack$SlackDateTime, tolerance = 30)
  expect_identical(slack$SlackType, rep(c("low", "high"), 4))

  monterey <- rtide::monterey
  monterey$DateTime <- dtt_add_hours(monterey$DateTime, 2L)
  slack2 <- tide_slack_data(monterey)
})

#' Is tide_harmonics
#'
#' Tests if object inherits from class tide_harmonics.
#'
#' @param x The object to test.
#' @export
is.tide_harmonics <- function(x) {
  inherits(x, "tide_harmonics")
}

check_tide_harmonics <- function(x) {
  if (!is.tide_harmonics(x)) stop("x is not class 'tide_harmonics'")

  if (!all(c("Station", "Node", "StationNode", "NodeYear") %in% names(x))) {
    stop("x is missing components", call. = FALSE)
  }

  check_data(x$Station,
    values = list(
      Station = "",
      Units = c("feet", "ft", "m", "metre"),
      Longitude = 1,
      Latitude = 1,
      Hours = c(-12, 12),
      TZ = "",
      Datum = 1
    ),
    nrow = c(1L, .Machine$integer.max),
    key = "Station"
  )

  check_data(x$Node,
    values = list(
      Node = "",
      Speed = 1
    ),
    nrow = c(1L, .Machine$integer.max),
    key = "Node"
  )

  if (!is.array(x$StationNode)) stop("StationNode must be an array", call. = FALSE)
  if (!is.array(x$NodeYear)) stop("NodeYear must be an array", call. = FALSE)
  if (mode(x$StationNode) != "numeric") {
    stop("StationNode must be a numeric array", call. = FALSE)
  }
  if (mode(x$NodeYear) != "numeric") {
    stop("NodeYear must be a numeric array", call. = FALSE)
  }

  if (!identical(dimnames(x$StationNode), list(x$Station$Station, x$Node$Node, c("A", "Kappa")))) {
    stop("StationNode has invalid dimnames", call. = FALSE)
  }

  if (!identical(dimnames(x$NodeYear)[c(1, 3)], list(x$Node$Node, c("NodeFactor", "EquilArg")))) {
    stop("NodeYear has invalid dimnames", call. = FALSE)
  }

  years <- dimnames(x$NodeYear)[2][[1]]
  years <- as.numeric(years)
  years <- diff(years)
  if (!all(years == 1)) stop("NodeYear has invalid dimnames", call. = FALSE)
  x
}

tide_harmonics <- function(x) {
  if (!is.list(x)) stop("x must be a list", call. = FALSE)

  if (!all(c(
    "name", "speed", "startyear", "equilarg", "nodefactor", "station",
    "units", "longitude", "latitude", "timezone", "tzfile", "datum",
    "A", "kappa"
  ) %in% names(x))) {
    stop("x missing components", call. = FALSE)
  }

  x$Station <- data.frame(
    Station = x$station, Units = x$unit, Longitude = x$longitude, Latitude = x$latitude,
    Hours = x$timezone, TZ = x$tzfile, Datum = x$datum, stringsAsFactors = FALSE
  )

  x$Station$Station <- enc2utf8(x$Station$Station)

  x$Node <- data.frame(Node = x$name, Speed = x$speed, stringsAsFactors = FALSE)
  x$StationNode <- abind::abind(A = x$A, Kappa = x$kappa, along = 3)
  dimnames(x$StationNode) <- list(x$Station$Station, x$Node$Node, c("A", "Kappa"))

  x$NodeYear <- abind::abind(NodeFactor = x$nodefactor, EquilArg = x$equilarg, along = 3)
  dimnames(x$NodeYear) <- list(
    x$Node$Node, seq(x$startyear, length.out = dim(x$NodeYear)[2]),
    c("NodeFactor", "EquilArg")
  )

  x <- x[c("Station", "Node", "StationNode", "NodeYear")]

  station <- order(x$Station$Station)
  x$Station <- x$Station[station, , drop = FALSE]
  x$StationNode <- x$StationNode[station, , , drop = FALSE]

  node <- order(x$Node$Node)
  x$Node <- x$Node[node, , drop = FALSE]
  x$StationNode <- x$StationNode[, node, , drop = FALSE]
  x$NodeYear <- x$NodeYear[node, , , drop = FALSE]

  x$Node <- tibble::as_tibble(x$Node)
  x$Station <- tibble::as_tibble(x$Station)

  class(x) <- c("tide_harmonics")
  check_tide_harmonics(x)
  x
}

#' @export
subset.tide_harmonics <- function(x, stations, ...) {
  stations <- tide_stations(stations, x)
  stations <- which(x$Station$Station %in% stations)
  x$Station <- x$Station[stations, , drop = FALSE]
  x$StationNode <- x$StationNode[stations, , , drop = FALSE]
  x
}

#' @export
format.tide_harmonics <- function(x, ...) {
  utils::str(x, ...)
}

#' @export
print.tide_harmonics <- function(x, ...) {
  cat(format(x, ...), "\n")
}

years_tide_harmonics <- function(x) {
  x <- dimnames(x$NodeYear)[[2]]
  x <- as.integer(as.character(x))
  x
}

test_that("harmonics correct", {
  expect_identical(anyDuplicated(years_tide_harmonics(rtide::harmonics)), 0L)
})

test_that("subset works", {
  h <- rtide::harmonics
  expect_equal(
    subset(h, h$Station$Station[2])$Station,
    h$Station[2, , drop = FALSE]
  )
})

test_that("years_tide_harmonics works", {
  expect_type(years_tide_harmonics(rtide::harmonics), "integer")
})

datetime2seconds <- function(x) {
  as.numeric(x)
}

has_name <- function(x, colname) colname %in% colnames(x)

seconds2datetime <- function(x) {
  as.POSIXct(x, origin = ISOdate(1970, 1, 1, 0), tz = "UTC")
}

test_that("datetimes and seconds work", {
  start <- ISOdate(1970, 1, 1, 0)
  expect_identical(datetime2seconds(seconds2datetime(datetime2seconds(start))), 0)
  expect_identical(datetime2seconds(seconds2datetime(datetime2seconds(dttr2::dtt_add_seconds(start, 3L)))), 3)
})

#' Harmonics
#'
#' A object of class tide_harmonics providing tidal harmonic data for US stations.
#'
#' Converted from harmonics-dwf-20151227-free, NOAA web site data processed by David Flater for XTide.
"harmonics"

#' Monterey Tide Height Data
#'
#' High/Low Tide Predictions from <https://tidesandcurrents.noaa.gov/tide_predictions.html>.
#'
#' @format A tbl data frame:
#' \describe{
#'   \item{Station}{The station name (chr).}
#'   \item{DateTime}{The date time (time).}
#'   \item{MLLW}{The tide height in m (dbl).}
#' }
"monterey"

#' Brandywine Tide Height Data
#'
#' High/Low Tide Predictions from <https://tidesandcurrents.noaa.gov/tide_predictions.html>.
#'
#' @format A tbl data frame:
#' \describe{
#'   \item{Station}{The station name (chr).}
#'   \item{DateTime}{The date time (time).}
#'   \item{MLLW}{The tide height in m (dbl).}
#' }
"brandywine"

ft2m <- function(x) {
  x * 0.3048
}

#' Tide Stations
#'
#' Gets vector of matching stations.
#'
#' @param stations A character vector of stations to match - treated as regular expressions.
#' @param harmonics The harmonics object.
#' @export
tide_stations <- function(stations = ".*", harmonics = rtide::harmonics) {
  chk_s3_class(stations, "character")
  chk_not_empty(stations)
  chk_not_any_na(stations)
  check_tide_harmonics(harmonics)
  if (!is.tide_harmonics(harmonics)) {
    stop("harmonics must be an object of class 'tide_harmonics'", call. = FALSE)
  }

  stations <- gsub("[(]", "[(]", stations)
  stations <- gsub("[)]", "[)]", stations)
  stations <- paste0("(", paste(stations, collapse = ")|("), ")")
  match <- grepl(stations, harmonics$Station$Station)
  match <- which(match)
  if (!length(match)) stop("no matching stations", call. = FALSE)
  harmonics$Station$Station[sort(unique(match))]
}

#' Tide Date Times
#'
#' Generates sequence of date times.
#'
#' @param minutes An integer of the number of minutes between tide heights
#' @param from A Date of the start of the period of interest
#' @param to A Date of the end of the period of interest
#' @param tz A string of the time zone.
#'
#' @return A POSIXct vector.
#' @export
#'
#' @examples
#' tide_datetimes()
tide_datetimes <- function(minutes = 60L, from = as.Date("2015-01-01"), to = as.Date("2015-12-31"),
                           tz = "America/Los_Angeles") {
  chk_number(minutes)
  chk_range(minutes, c(1, 60))

  if (inherits(minutes, "numeric")) {
    if (minutes %% 1 != 0) { # If modulo isn't 0, decimal value is present
      warning("Truncating minutes interval to whole number", call. = FALSE)
    }
    minutes <- as.integer(minutes)
  }

  chk_date(from)
  chk_date(to)
  chk_tz(tz)

  from <- ISOdatetime(
    year = dtt_year(from), month = dtt_month(from),
    day = dtt_day(from), hour = 0, min = 0, sec = 0, tz = tz
  )
  to <- ISOdatetime(
    year = dtt_year(to), month = dtt_month(to),
    day = dtt_day(to), hour = 23, min = 59, sec = 59, tz = tz
  )

  seq(from, to, by = paste(minutes, "min"))
}

hours_year <- function(datetime) {
  chk_s3_class(datetime, "POSIXct")
  chk_not_any_na(datetime)
  stopifnot(identical(dtt_tz(datetime), "UTC"))

  year <- dtt_year(datetime)

  startdatetime <- ISOdate(year, 1, 1, 0, tz = "UTC")
  hours <- difftime(datetime, startdatetime, units = "hours")
  hours <- as.numeric(hours)
  hours
}

tide_height_datetime <- function(d, h) {
  h$NodeYear <- h$NodeYear[, as.character(dtt_year(d)), , drop = FALSE]

  height <- h$Station$Datum + sum(h$NodeYear[, , "NodeFactor"] * h$StationNode[, , "A"] *
    cos((h$Node$Speed * (hours_year(d) - h$Station$Hours) +
      h$NodeYear[, , "EquilArg"] - h$StationNode[, , "Kappa"]) * pi / 180))

  height
}

tide_height_data_datetime <- function(d, h) {
  d$TideHeight <- tide_height_datetime(d$DateTime, h)
  d
}

tide_height_data_station <- function(data, harmonics) {
  harmonics <- subset(harmonics, paste0("^", data$Station[1], "$"))
  data <- split(data, seq_len(nrow(data)))
  data <- lapply(data, FUN = tide_height_data_datetime, h = harmonics)
  data <- do.call("rbind", data)
  if (harmonics$Station$Units %in% c("feet", "ft")) {
    data$TideHeight <- ft2m(data$TideHeight)
  }
  data
}


#' Tide Height Data
#'
#' Calculates tide height at specified stations at particular date times based on the supplied harmonics object.
#'
#' @param data A data frame with the columns Station and DateTime.
#' @inheritParams tide_stations
#' @return A data frame of the tide heights in m.
#' @export
tide_height_data <- function(data, harmonics = rtide::harmonics) {
  check_data(data,
    values = list(Station = "", DateTime = Sys.time()),
    nrow = c(1L, .Machine$integer.max)
  )

  if (!all(data$Station %in% tide_stations(harmonics = harmonics))) {
    stop("unrecognised stations", call. = FALSE)
  }

  if (has_name(data, "TideHeight")) {
    stop("data already has 'TideHeight' column", call. = FALSE)
  }

  tz <- dtt_tz(data$DateTime)
  data$DateTime <- dtt_adjust_tz(data$DateTime, tz = "UTC")

  years <- range(dtt_year(data$DateTime), na.rm = TRUE)
  if (!all(years %in% years_tide_harmonics(harmonics))) {
    stop("years are outside harmonics range", call. = FALSE)
  }

  station <- data$Station
  data <- split(data, station)
  data <- lapply(data, FUN = tide_height_data_station, harmonics = harmonics)
  op <- options(stringsAsFactors = FALSE)
  on.exit(options(op))
  data <- do.call("rbind", data)

  data$DateTime <- dtt_adjust_tz(data$DateTime, tz = tz)
  data <- data[order(data$Station, data$DateTime), ]
  tibble::as_tibble(data)
}

#' Tide Height
#'
#' Calculates tide height at specified stations based on the supplied harmonics object.
#'
#' @inheritParams tide_stations
#' @inheritParams tide_datetimes
#' @return A data frame of the tide heights in m by the number of minutes for each station from from to to.
#' @export
tide_height <- function(stations = "Monterey Harbor", minutes = 60L,
                        from = as.Date("2015-01-01"), to = as.Date("2015-01-01"), tz = "UTC",
                        harmonics = rtide::harmonics) {
  stations <- tide_stations(stations, harmonics)
  datetimes <- tide_datetimes(minutes = minutes, from = from, to = to, tz = tz)

  data <- expand.grid(Station = stations, DateTime = datetimes, stringsAsFactors = FALSE)
  tide_height_data(data, harmonics = harmonics)
}

test_that("tide_height works", {
  expect_error(chk::check_data(tide_height(),
    values = list(
      Station = "", DateTime = Sys.time(), TideHeight = 1
    ),
    nrow = 24,
    key = "DateTime",
    exclusive = TRUE,
    order = TRUE
  ), NA)

  expect_s3_class(tide_height(stations = ".*"), "data.frame")
})

test_that("hours_year works", {
  expect_error(hours_year(1))
  expect_identical(hours_year(ISOdate(2000, 1, 1, 1, tz = "UTC")), 1)
  expect_identical(hours_year(ISOdate(2000, 1, 1, c(1, 3, 1), tz = "UTC")), c(1L, 3, 1))
  expect_identical(hours_year(ISOdate(2001, 1, 1, c(1, 3, 1), tz = "UTC")), c(1, 3, 1))
  expect_identical(hours_year(ISOdate(2001, 1, 2, c(1, 3, 1), tz = "UTC")), c(25, 27, 25))
})

test_that("tide_datetimes works", {
  minutes <- 17L
  from <- as.Date("2000-02-01")
  to <- as.Date("2000-05-02")
  tz <- "America/Los_Angeles"

  expect_equal(dtt_tz(tide_datetimes(minutes = minutes, from = from, to = to, tz = tz)), "America/Los_Angeles")
  expect_equal(tide_datetimes(minutes = minutes, from = from, to = to, tz = tz)[1], ISOdate(2000, 02, 01, 0, tz = tz))
  expect_equal(max(dtt_date(tide_datetimes(minutes = minutes, from = from, to = to, tz = tz))), to)

  expect_error(dtt_tz(tide_datetimes(minutes = minutes, from = to, to = from, tz = tz)))

  expect_identical(tide_datetimes(minutes = 1), tide_datetimes(minutes = 1L))
  expect_warning(tide_datetimes(minutes = 1.9), "Truncating minutes interval to whole number")
  expect_chk_error(tide_datetimes(minutes = 0L), "^`minutes` must be between 1 and 60, not 0[.]$")
  expect_chk_error(tide_datetimes(minutes = 0.9), "^`minutes` must be between 1 and 60, not 0[.]9[.]$")
})

test_that("tide_height_data works", {
  data <- data.frame(
    Station = "Monterey, Monterey Harbor, California",
    DateTime = ISOdate(2015, 1, 1, 10, tz = "America/Los_Angeles"),
    stringsAsFactors = FALSE
  )

  expect_error(chk::check_data(tide_height_data(data),
    values = list(
      Station = "", DateTime = Sys.time(), TideHeight = 1
    ),
    nrow = 1, exclusive = TRUE, order = TRUE
  ), NA)
  expect_identical(dtt_tz(data$DateTime), "America/Los_Angeles")
})

test_that("tide_height_data predictions", {
  expect_equal(rtide::monterey$MLLW,
    tide_height_data(rtide::monterey)$TideHeight,
    tolerance = 0.002
  )
  expect_equal(rtide::brandywine$MLLW,
    tide_height_data(rtide::brandywine)$TideHeight,
    tolerance = 0.002
  )
})

test_that("tide_height_data checks", {
  data <- data.frame(
    Station = "Monterey, Monterey Harbor, California",
    DateTime = ISOdate(2015, 1, 1, 10, tz = "America/Los_Angeles"),
    stringsAsFactors = FALSE
  )

  data$TideHeight <- 1

  expect_error(tide_height_data(data), "data already has 'TideHeight' column")

  data$TideHeight <- NULL
  dtt_year(data$DateTime) <- 1699L
  expect_error(tide_height_data(data), "years are outside harmonics range")
})

test_that("tide_height_data tz", {
  data <- data.frame(
    Station = "Monterey, Monterey Harbor, California",
    DateTime = ISOdate(2015, 1, 1, 10, tz = "America/Los_Angeles"),
    stringsAsFactors = FALSE
  )

  data2 <- data
  data2$DateTime <- dtt_adjust_tz(data2$DateTime, tz = "America/New_York") # Formerly EST

  expect_equal(tide_height_data(data)$TideHeight, tide_height_data(data2)$TideHeight)
})

test_that("tide_stations works", {
  expect_equal(tide_stations("Monterey", rtide::harmonics), c("Elkhorn Slough railroad bridge, Monterey Bay, California", "Monterey, Monterey Harbor, California"))
  expect_equal(
    tide_stations(
      c("Elkhorn Slough railroad bridge, Monterey Bay, California", "Monterey, Monterey Harbor, California")
    ),
    c("Elkhorn Slough railroad bridge, Monterey Bay, California", "Monterey, Monterey Harbor, California")
  )

  expect_equal(tide_stations("Annapolis (US Naval Academy), Severn River, Maryland", rtide::harmonics), "Annapolis (US Naval Academy), Severn River, Maryland")

  expect_error(tide_stations("^Monterey$", rtide::harmonics), "no matching stations")
  expect_chk_error(tide_stations(1, rtide::harmonics), "^`stations` must inherit from S3 class 'character'")
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtide-package.R
\docType{package}
\name{rtide-package}
\alias{rtide}
\alias{rtide-package}
\title{rtide: Tide Heights}
\description{
Calculates tide heights based on tide station harmonics. It includes the harmonics data for 637 US stations. The harmonics data was converted from \url{https://github.com/poissonconsulting/rtide/blob/main/data-raw/harmonics-dwf-20151227-free.tar.bz2}, NOAA web site data processed by David Flater for 'XTide'. The code to calculate tide heights from the harmonics is based on 'XTide'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/millerlp/rtide}
  \item Report bugs at \url{https://github.com/millerlp/rtide/issues}
}

}
\author{
\strong{Maintainer}: Luke Miller \email{contact@lukemiller.org}

Authors:
\itemize{
  \item Luke Miller \email{contact@lukemiller.org}
  \item Joe Thorley \email{joe@poissonconsulting.ca} (\href{https://orcid.org/0000-0002-7683-4592}{ORCID})
  \item Abram Fleishman
}

Other contributors:
\itemize{
  \item Poisson Consulting [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tide-height.R
\name{tide_datetimes}
\alias{tide_datetimes}
\title{Tide Date Times}
\usage{
tide_datetimes(
  minutes = 60L,
  from = as.Date("2015-01-01"),
  to = as.Date("2015-12-31"),
  tz = "America/Los_Angeles"
)
}
\arguments{
\item{minutes}{An integer of the number of minutes between tide heights}

\item{from}{A Date of the start of the period of interest}

\item{to}{A Date of the end of the period of interest}

\item{tz}{A string of the time zone.}
}
\value{
A POSIXct vector.
}
\description{
Generates sequence of date times.
}
\examples{
tide_datetimes()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{harmonics}
\alias{harmonics}
\title{Harmonics}
\format{
An object of class \code{tide_harmonics} of length 4.
}
\usage{
harmonics
}
\description{
A object of class tide_harmonics providing tidal harmonic data for US stations.
}
\details{
Converted from harmonics-dwf-20151227-free, NOAA web site data processed by David Flater for XTide.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tide-slack.R
\name{tide_slack_data}
\alias{tide_slack_data}
\title{Tide Slack Data}
\usage{
tide_slack_data(data, harmonics = rtide::harmonics)
}
\arguments{
\item{data}{A data frame with the columns Station and DateTime.}

\item{harmonics}{The harmonics object.}
}
\value{
A data frame of the slack tide date times and heights in m.
}
\description{
Determines the closest slack tide for specified stations at particular date times based on the supplied harmonics object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tide-height.R
\name{tide_stations}
\alias{tide_stations}
\title{Tide Stations}
\usage{
tide_stations(stations = ".*", harmonics = rtide::harmonics)
}
\arguments{
\item{stations}{A character vector of stations to match - treated as regular expressions.}

\item{harmonics}{The harmonics object.}
}
\description{
Gets vector of matching stations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tide-height.R
\name{tide_height}
\alias{tide_height}
\title{Tide Height}
\usage{
tide_height(
  stations = "Monterey Harbor",
  minutes = 60L,
  from = as.Date("2015-01-01"),
  to = as.Date("2015-01-01"),
  tz = "UTC",
  harmonics = rtide::harmonics
)
}
\arguments{
\item{stations}{A character vector of stations to match - treated as regular expressions.}

\item{minutes}{An integer of the number of minutes between tide heights}

\item{from}{A Date of the start of the period of interest}

\item{to}{A Date of the end of the period of interest}

\item{tz}{A string of the time zone.}

\item{harmonics}{The harmonics object.}
}
\value{
A data frame of the tide heights in m by the number of minutes for each station from from to to.
}
\description{
Calculates tide height at specified stations based on the supplied harmonics object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tide-harmonics.R
\name{is.tide_harmonics}
\alias{is.tide_harmonics}
\title{Is tide_harmonics}
\usage{
is.tide_harmonics(x)
}
\arguments{
\item{x}{The object to test.}
}
\description{
Tests if object inherits from class tide_harmonics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{monterey}
\alias{monterey}
\title{Monterey Tide Height Data}
\format{
A tbl data frame:
\describe{
\item{Station}{The station name (chr).}
\item{DateTime}{The date time (time).}
\item{MLLW}{The tide height in m (dbl).}
}
}
\usage{
monterey
}
\description{
High/Low Tide Predictions from \url{https://tidesandcurrents.noaa.gov/tide_predictions.html}.
}
\keyword{datasets}
